@@ -4,28 +4,60 @@ from bson import ObjectId, SON
 from bson.dbref import DBRef
 import pymongo
 import six
+from six import iteritems
 
 from mongoengine.base import UPDATE_OPERATORS
 from mongoengine.common import _import_class
-from mongoengine.connection import get_connection
 from mongoengine.errors import InvalidQueryError
-from mongoengine.python_support import IS_PYMONGO_3
 
-__all__ = ('query', 'update')
+__all__ = ("query", "update")
 
-COMPARISON_OPERATORS = ('ne', 'gt', 'gte', 'lt', 'lte', 'in', 'nin', 'mod',
-                        'all', 'size', 'exists', 'not', 'elemMatch', 'type')
-GEO_OPERATORS = ('within_distance', 'within_spherical_distance',
-                 'within_box', 'within_polygon', 'near', 'near_sphere',
-                 'max_distance', 'min_distance', 'geo_within', 'geo_within_box',
-                 'geo_within_polygon', 'geo_within_center',
-                 'geo_within_sphere', 'geo_intersects')
-STRING_OPERATORS = ('contains', 'icontains', 'startswith',
-                    'istartswith', 'endswith', 'iendswith',
-                    'exact', 'iexact')
-CUSTOM_OPERATORS = ('match',)
-MATCH_OPERATORS = (COMPARISON_OPERATORS + GEO_OPERATORS +
-                   STRING_OPERATORS + CUSTOM_OPERATORS)
+COMPARISON_OPERATORS = (
+    "ne",
+    "gt",
+    "gte",
+    "lt",
+    "lte",
+    "in",
+    "nin",
+    "mod",
+    "all",
+    "size",
+    "exists",
+    "not",
+    "elemMatch",
+    "type",
+)
+GEO_OPERATORS = (
+    "within_distance",
+    "within_spherical_distance",
+    "within_box",
+    "within_polygon",
+    "near",
+    "near_sphere",
+    "max_distance",
+    "min_distance",
+    "geo_within",
+    "geo_within_box",
+    "geo_within_polygon",
+    "geo_within_center",
+    "geo_within_sphere",
+    "geo_intersects",
+)
+STRING_OPERATORS = (
+    "contains",
+    "icontains",
+    "startswith",
+    "istartswith",
+    "endswith",
+    "iendswith",
+    "exact",
+    "iexact",
+)
+CUSTOM_OPERATORS = ("match",)
+MATCH_OPERATORS = (
+    COMPARISON_OPERATORS + GEO_OPERATORS + STRING_OPERATORS + CUSTOM_OPERATORS
+)
 
 
 # TODO make this less complex
@@ -34,11 +66,11 @@ def query(_doc_cls=None, **kwargs):
     mongo_query = {}
     merge_query = defaultdict(list)
     for key, value in sorted(kwargs.items()):
-        if key == '__raw__':
+        if key == "__raw__":
             mongo_query.update(value)
             continue
 
-        parts = key.rsplit('__')
+        parts = key.rsplit("__")
         indices = [(i, p) for i, p in enumerate(parts) if p.isdigit()]
         parts = [part for part in parts if not part.isdigit()]
         # Check for an operator and transform to mongo-style if there is
@@ -47,11 +79,11 @@ def query(_doc_cls=None, **kwargs):
             op = parts.pop()
 
         # Allow to escape operator-like field name by __
-        if len(parts) > 1 and parts[-1] == '':
+        if len(parts) > 1 and parts[-1] == "":
             parts.pop()
 
         negate = False
-        if len(parts) > 1 and parts[-1] == 'not':
+        if len(parts) > 1 and parts[-1] == "not":
             parts.pop()
             negate = True
 
@@ -63,8 +95,8 @@ def query(_doc_cls=None, **kwargs):
                 raise InvalidQueryError(e)
             parts = []
 
-            CachedReferenceField = _import_class('CachedReferenceField')
-            GenericReferenceField = _import_class('GenericReferenceField')
+            CachedReferenceField = _import_class("CachedReferenceField")
+            GenericReferenceField = _import_class("GenericReferenceField")
 
             cleaned_fields = []
             for field in fields:
@@ -74,7 +106,7 @@ def query(_doc_cls=None, **kwargs):
                     append_field = False
                 # is last and CachedReferenceField
                 elif isinstance(field, CachedReferenceField) and fields[-1] == field:
-                    parts.append('%s._id' % field.db_field)
+                    parts.append("%s._id" % field.db_field)
                 else:
                     parts.append(field.db_field)
 
@@ -84,23 +116,15 @@ def query(_doc_cls=None, **kwargs):
             # Convert value to proper value
             field = cleaned_fields[-1]
 
-            singular_ops = [None, 'ne', 'gt', 'gte', 'lt', 'lte', 'not']
+            singular_ops = [None, "ne", "gt", "gte", "lt", "lte", "not"]
             singular_ops += STRING_OPERATORS
             if op in singular_ops:
-                if isinstance(field, six.string_types):
-                    if (op in STRING_OPERATORS and
-                            isinstance(value, six.string_types)):
-                        StringField = _import_class('StringField')
-                        value = StringField.prepare_query_value(op, value)
-                    else:
-                        value = field
-                else:
-                    value = field.prepare_query_value(op, value)
+                value = field.prepare_query_value(op, value)
 
-                    if isinstance(field, CachedReferenceField) and value:
-                        value = value['_id']
+                if isinstance(field, CachedReferenceField) and value:
+                    value = value["_id"]
 
-            elif op in ('in', 'nin', 'all', 'near') and not isinstance(value, dict):
+            elif op in ("in", "nin", "all", "near") and not isinstance(value, dict):
                 # Raise an error if the in/nin/all/near param is not iterable.
                 value = _prepare_query_for_iterable(field, op, value)
 
@@ -110,73 +134,77 @@ def query(_doc_cls=None, **kwargs):
             # * If the value is an ObjectId, the key should be "field_name._ref.$id".
             if isinstance(field, GenericReferenceField):
                 if isinstance(value, DBRef):
-                    parts[-1] += '._ref'
+                    parts[-1] += "._ref"
                 elif isinstance(value, ObjectId):
-                    parts[-1] += '._ref.$id'
+                    parts[-1] += "._ref.$id"
 
         # if op and op not in COMPARISON_OPERATORS:
         if op:
             if op in GEO_OPERATORS:
                 value = _geo_operator(field, op, value)
-            elif op in ('match', 'elemMatch'):
-                ListField = _import_class('ListField')
-                EmbeddedDocumentField = _import_class('EmbeddedDocumentField')
+            elif op in ("match", "elemMatch"):
+                ListField = _import_class("ListField")
+                EmbeddedDocumentField = _import_class("EmbeddedDocumentField")
                 if (
-                    isinstance(value, dict) and
-                    isinstance(field, ListField) and
-                    isinstance(field.field, EmbeddedDocumentField)
+                    isinstance(value, dict)
+                    and isinstance(field, ListField)
+                    and isinstance(field.field, EmbeddedDocumentField)
                 ):
                     value = query(field.field.document_type, **value)
                 else:
                     value = field.prepare_query_value(op, value)
-                value = {'$elemMatch': value}
+                value = {"$elemMatch": value}
             elif op in CUSTOM_OPERATORS:
-                NotImplementedError('Custom method "%s" has not '
-                                    'been implemented' % op)
+                NotImplementedError(
+                    'Custom method "%s" has not ' "been implemented" % op
+                )
             elif op not in STRING_OPERATORS:
-                value = {'$' + op: value}
+                value = {"$" + op: value}
 
         if negate:
-            value = {'$not': value}
+            value = {"$not": value}
 
         for i, part in indices:
             parts.insert(i, part)
 
-        key = '.'.join(parts)
+        key = ".".join(parts)
 
         if op is None or key not in mongo_query:
             mongo_query[key] = value
         elif key in mongo_query:
-            if isinstance(mongo_query[key], dict):
+            if isinstance(mongo_query[key], dict) and isinstance(value, dict):
                 mongo_query[key].update(value)
                 # $max/minDistance needs to come last - convert to SON
                 value_dict = mongo_query[key]
-                if ('$maxDistance' in value_dict or '$minDistance' in value_dict) and \
-                        ('$near' in value_dict or '$nearSphere' in value_dict):
+                if ("$maxDistance" in value_dict or "$minDistance" in value_dict) and (
+                    "$near" in value_dict or "$nearSphere" in value_dict
+                ):
                     value_son = SON()
-                    for k, v in value_dict.iteritems():
-                        if k == '$maxDistance' or k == '$minDistance':
+                    for k, v in iteritems(value_dict):
+                        if k == "$maxDistance" or k == "$minDistance":
                             continue
                         value_son[k] = v
                     # Required for MongoDB >= 2.6, may fail when combining
                     # PyMongo 3+ and MongoDB < 2.6
                     near_embedded = False
-                    for near_op in ('$near', '$nearSphere'):
-                        if isinstance(value_dict.get(near_op), dict) and (
-                                IS_PYMONGO_3 or get_connection().max_wire_version > 1):
+                    for near_op in ("$near", "$nearSphere"):
+                        if isinstance(value_dict.get(near_op), dict):
                             value_son[near_op] = SON(value_son[near_op])
-                            if '$maxDistance' in value_dict:
-                                value_son[near_op][
-                                    '$maxDistance'] = value_dict['$maxDistance']
-                            if '$minDistance' in value_dict:
-                                value_son[near_op][
-                                    '$minDistance'] = value_dict['$minDistance']
+                            if "$maxDistance" in value_dict:
+                                value_son[near_op]["$maxDistance"] = value_dict[
+                                    "$maxDistance"
+                                ]
+                            if "$minDistance" in value_dict:
+                                value_son[near_op]["$minDistance"] = value_dict[
+                                    "$minDistance"
+                                ]
                             near_embedded = True
+
                     if not near_embedded:
-                        if '$maxDistance' in value_dict:
-                            value_son['$maxDistance'] = value_dict['$maxDistance']
-                        if '$minDistance' in value_dict:
-                            value_son['$minDistance'] = value_dict['$minDistance']
+                        if "$maxDistance" in value_dict:
+                            value_son["$maxDistance"] = value_dict["$maxDistance"]
+                        if "$minDistance" in value_dict:
+                            value_son["$minDistance"] = value_dict["$minDistance"]
                     mongo_query[key] = value_son
             else:
                 # Store for manually merging later
@@ -188,10 +216,10 @@ def query(_doc_cls=None, **kwargs):
         del mongo_query[k]
         if isinstance(v, list):
             value = [{k: val} for val in v]
-            if '$and' in mongo_query.keys():
-                mongo_query['$and'].extend(value)
+            if "$and" in mongo_query.keys():
+                mongo_query["$and"].extend(value)
             else:
-                mongo_query['$and'] = value
+                mongo_query["$and"] = value
 
     return mongo_query
 
@@ -201,37 +229,44 @@ def update(_doc_cls=None, **update):
     format.
     """
     mongo_update = {}
+
     for key, value in update.items():
-        if key == '__raw__':
+        if key == "__raw__":
             mongo_update.update(value)
             continue
-        parts = key.split('__')
+
+        parts = key.split("__")
+
         # if there is no operator, default to 'set'
         if len(parts) < 3 and parts[0] not in UPDATE_OPERATORS:
-            parts.insert(0, 'set')
+            parts.insert(0, "set")
+
         # Check for an operator and transform to mongo-style if there is
         op = None
         if parts[0] in UPDATE_OPERATORS:
             op = parts.pop(0)
             # Convert Pythonic names to Mongo equivalents
-            if op in ('push_all', 'pull_all'):
-                op = op.replace('_all', 'All')
-            elif op == 'dec':
+            operator_map = {
+                "push_all": "pushAll",
+                "pull_all": "pullAll",
+                "dec": "inc",
+                "add_to_set": "addToSet",
+                "set_on_insert": "setOnInsert",
+            }
+            if op == "dec":
                 # Support decrement by flipping a positive value's sign
                 # and using 'inc'
-                op = 'inc'
                 value = -value
-            elif op == 'add_to_set':
-                op = 'addToSet'
-            elif op == 'set_on_insert':
-                op = 'setOnInsert'
+            # If the operator doesn't found from operator map, the op value
+            # will stay unchanged
+            op = operator_map.get(op, op)
 
         match = None
         if parts[-1] in COMPARISON_OPERATORS:
             match = parts.pop()
 
         # Allow to escape operator-like field name by __
-        if len(parts) > 1 and parts[-1] == '':
+        if len(parts) > 1 and parts[-1] == "":
             parts.pop()
 
         if _doc_cls:
@@ -248,8 +283,8 @@ def update(_doc_cls=None, **update):
                 append_field = True
                 if isinstance(field, six.string_types):
                     # Convert the S operator to $
-                    if field == 'S':
-                        field = '$'
+                    if field == "S":
+                        field = "$"
                     parts.append(field)
                     append_field = False
                 else:
@@ -257,7 +292,7 @@ def update(_doc_cls=None, **update):
                 if append_field:
                     appended_sub_field = False
                     cleaned_fields.append(field)
-                    if hasattr(field, 'field'):
+                    if hasattr(field, "field"):
                         cleaned_fields.append(field.field)
                         appended_sub_field = True
 
@@ -267,54 +302,53 @@ def update(_doc_cls=None, **update):
             else:
                 field = cleaned_fields[-1]
 
-            GeoJsonBaseField = _import_class('GeoJsonBaseField')
+            GeoJsonBaseField = _import_class("GeoJsonBaseField")
             if isinstance(field, GeoJsonBaseField):
                 value = field.to_mongo(value)
 
-            if op == 'pull':
+            if op == "pull":
                 if field.required or value is not None:
-                    if match == 'in' and not isinstance(value, dict):
+                    if match in ("in", "nin") and not isinstance(value, dict):
                         value = _prepare_query_for_iterable(field, op, value)
                     else:
                         value = field.prepare_query_value(op, value)
-            elif op == 'push' and isinstance(value, (list, tuple, set)):
+            elif op == "push" and isinstance(value, (list, tuple, set)):
                 value = [field.prepare_query_value(op, v) for v in value]
-            elif op in (None, 'set', 'push'):
+            elif op in (None, "set", "push"):
                 if field.required or value is not None:
                     value = field.prepare_query_value(op, value)
-            elif op in ('pushAll', 'pullAll'):
+            elif op in ("pushAll", "pullAll"):
                 value = [field.prepare_query_value(op, v) for v in value]
-            elif op in ('addToSet', 'setOnInsert'):
+            elif op in ("addToSet", "setOnInsert"):
                 if isinstance(value, (list, tuple, set)):
                     value = [field.prepare_query_value(op, v) for v in value]
                 elif field.required or value is not None:
                     value = field.prepare_query_value(op, value)
-            elif op == 'unset':
+            elif op == "unset":
                 value = 1
+            elif op == "inc":
+                value = field.prepare_query_value(op, value)
 
         if match:
-            match = '$' + match
+            match = "$" + match
             value = {match: value}
 
-        key = '.'.join(parts)
+        key = ".".join(parts)
 
-        if not op:
-            raise InvalidQueryError('Updates must supply an operation '
-                                    'eg: set__FIELD=value')
-
-        if 'pull' in op and '.' in key:
+        if "pull" in op and "." in key:
             # Dot operators don't work on pull operations
             # unless they point to a list field
             # Otherwise it uses nested dict syntax
-            if op == 'pullAll':
-                raise InvalidQueryError('pullAll operations only support '
-                                        'a single field depth')
+            if op == "pullAll":
+                raise InvalidQueryError(
+                    "pullAll operations only support a single field depth"
+                )
 
             # Look for the last list field and use dot notation until there
             field_classes = [c.__class__ for c in cleaned_fields]
             field_classes.reverse()
-            ListField = _import_class('ListField')
-            EmbeddedDocumentListField = _import_class('EmbeddedDocumentListField')
+            ListField = _import_class("ListField")
+            EmbeddedDocumentListField = _import_class("EmbeddedDocumentListField")
             if ListField in field_classes or EmbeddedDocumentListField in field_classes:
                 # Join all fields via dot notation to the last ListField or EmbeddedDocumentListField
                 # Then process as normal
@@ -323,37 +357,36 @@ def update(_doc_cls=None, **update):
                 else:
                     _check_field = EmbeddedDocumentListField
 
-                last_listField = len(
-                    cleaned_fields) - field_classes.index(_check_field)
-                key = '.'.join(parts[:last_listField])
+                last_listField = len(cleaned_fields) - field_classes.index(_check_field)
+                key = ".".join(parts[:last_listField])
                 parts = parts[last_listField:]
                 parts.insert(0, key)
 
             parts.reverse()
             for key in parts:
                 value = {key: value}
-        elif op == 'addToSet' and isinstance(value, list):
-            value = {key: {'$each': value}}
-        elif op in ('push', 'pushAll'):
+        elif op == "addToSet" and isinstance(value, list):
+            value = {key: {"$each": value}}
+        elif op in ("push", "pushAll"):
             if parts[-1].isdigit():
-                key = parts[0]
+                key = ".".join(parts[0:-1])
                 position = int(parts[-1])
                 # $position expects an iterable. If pushing a single value,
                 # wrap it in a list.
                 if not isinstance(value, (set, tuple, list)):
                     value = [value]
-                value = {key: {'$each': value, '$position': position}}
+                value = {key: {"$each": value, "$position": position}}
             else:
-                if op == 'pushAll':
-                    op = 'push'  # convert to non-deprecated keyword
+                if op == "pushAll":
+                    op = "push"  # convert to non-deprecated keyword
                     if not isinstance(value, (set, tuple, list)):
                         value = [value]
-                    value = {key: {'$each': value}}
+                    value = {key: {"$each": value}}
                 else:
                     value = {key: value}
         else:
             value = {key: value}
-        key = '$' + op
+        key = "$" + op
         if key not in mongo_update:
             mongo_update[key] = value
         elif key in mongo_update and isinstance(mongo_update[key], dict):
@@ -364,45 +397,45 @@ def update(_doc_cls=None, **update):
 
 def _geo_operator(field, op, value):
     """Helper to return the query for a given geo query."""
-    if op == 'max_distance':
-        value = {'$maxDistance': value}
-    elif op == 'min_distance':
-        value = {'$minDistance': value}
+    if op == "max_distance":
+        value = {"$maxDistance": value}
+    elif op == "min_distance":
+        value = {"$minDistance": value}
     elif field._geo_index == pymongo.GEO2D:
-        if op == 'within_distance':
-            value = {'$within': {'$center': value}}
-        elif op == 'within_spherical_distance':
-            value = {'$within': {'$centerSphere': value}}
-        elif op == 'within_polygon':
-            value = {'$within': {'$polygon': value}}
-        elif op == 'near':
-            value = {'$near': value}
-        elif op == 'near_sphere':
-            value = {'$nearSphere': value}
-        elif op == 'within_box':
-            value = {'$within': {'$box': value}}
-        else:
-            raise NotImplementedError('Geo method "%s" has not been '
-                                      'implemented for a GeoPointField' % op)
-    else:
-        if op == 'geo_within':
-            value = {'$geoWithin': _infer_geometry(value)}
-        elif op == 'geo_within_box':
-            value = {'$geoWithin': {'$box': value}}
-        elif op == 'geo_within_polygon':
-            value = {'$geoWithin': {'$polygon': value}}
-        elif op == 'geo_within_center':
-            value = {'$geoWithin': {'$center': value}}
-        elif op == 'geo_within_sphere':
-            value = {'$geoWithin': {'$centerSphere': value}}
-        elif op == 'geo_intersects':
-            value = {'$geoIntersects': _infer_geometry(value)}
-        elif op == 'near':
-            value = {'$near': _infer_geometry(value)}
+        if op == "within_distance":
+            value = {"$within": {"$center": value}}
+        elif op == "within_spherical_distance":
+            value = {"$within": {"$centerSphere": value}}
+        elif op == "within_polygon":
+            value = {"$within": {"$polygon": value}}
+        elif op == "near":
+            value = {"$near": value}
+        elif op == "near_sphere":
+            value = {"$nearSphere": value}
+        elif op == "within_box":
+            value = {"$within": {"$box": value}}
         else:
             raise NotImplementedError(
-                'Geo method "%s" has not been implemented for a %s '
-                % (op, field._name)
+                'Geo method "%s" has not been ' "implemented for a GeoPointField" % op
+            )
+    else:
+        if op == "geo_within":
+            value = {"$geoWithin": _infer_geometry(value)}
+        elif op == "geo_within_box":
+            value = {"$geoWithin": {"$box": value}}
+        elif op == "geo_within_polygon":
+            value = {"$geoWithin": {"$polygon": value}}
+        elif op == "geo_within_center":
+            value = {"$geoWithin": {"$center": value}}
+        elif op == "geo_within_sphere":
+            value = {"$geoWithin": {"$centerSphere": value}}
+        elif op == "geo_intersects":
+            value = {"$geoIntersects": _infer_geometry(value)}
+        elif op == "near":
+            value = {"$near": _infer_geometry(value)}
+        else:
+            raise NotImplementedError(
+                'Geo method "%s" has not been implemented for a %s ' % (op, field._name)
             )
     return value
 
@@ -412,52 +445,58 @@ def _infer_geometry(value):
     given value.
     """
     if isinstance(value, dict):
-        if '$geometry' in value:
+        if "$geometry" in value:
             return value
-        elif 'coordinates' in value and 'type' in value:
-            return {'$geometry': value}
-        raise InvalidQueryError('Invalid $geometry dictionary should have '
-                                'type and coordinates keys')
+        elif "coordinates" in value and "type" in value:
+            return {"$geometry": value}
+        raise InvalidQueryError(
+            "Invalid $geometry dictionary should have type and coordinates keys"
+        )
     elif isinstance(value, (list, set)):
         # TODO: shouldn't we test value[0][0][0][0] to see if it is MultiPolygon?
-        # TODO: should both TypeError and IndexError be alike interpreted?
 
         try:
             value[0][0][0]
-            return {'$geometry': {'type': 'Polygon', 'coordinates': value}}
+            return {"$geometry": {"type": "Polygon", "coordinates": value}}
         except (TypeError, IndexError):
             pass
 
         try:
             value[0][0]
-            return {'$geometry': {'type': 'LineString', 'coordinates': value}}
+            return {"$geometry": {"type": "LineString", "coordinates": value}}
         except (TypeError, IndexError):
             pass
 
         try:
             value[0]
-            return {'$geometry': {'type': 'Point', 'coordinates': value}}
+            return {"$geometry": {"type": "Point", "coordinates": value}}
         except (TypeError, IndexError):
             pass
 
-    raise InvalidQueryError('Invalid $geometry data. Can be either a '
-                            'dictionary or (nested) lists of coordinate(s)')
+    raise InvalidQueryError(
+        "Invalid $geometry data. Can be either a "
+        "dictionary or (nested) lists of coordinate(s)"
+    )
 
 
 def _prepare_query_for_iterable(field, op, value):
     # We need a special check for BaseDocument, because - although it's iterable - using
     # it as such in the context of this method is most definitely a mistake.
-    BaseDocument = _import_class('BaseDocument')
+    BaseDocument = _import_class("BaseDocument")
 
     if isinstance(value, BaseDocument):
-        raise TypeError("When using the `in`, `nin`, `all`, or "
-                        "`near`-operators you can\'t use a "
-                        "`Document`, you must wrap your object "
-                        "in a list (object -> [object]).")
+        raise TypeError(
+            "When using the `in`, `nin`, `all`, or "
+            "`near`-operators you can't use a "
+            "`Document`, you must wrap your object "
+            "in a list (object -> [object])."
+        )
 
-    if not hasattr(value, '__iter__'):
-        raise TypeError("The `in`, `nin`, `all`, or "
-                        "`near`-operators must be applied to an "
-                        "iterable (e.g. a list).")
+    if not hasattr(value, "__iter__"):
+        raise TypeError(
+            "The `in`, `nin`, `all`, or "
+            "`near`-operators must be applied to an "
+            "iterable (e.g. a list)."
+        )
 
     return [field.prepare_query_value(op, v) for v in value]
